#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# input-remapper - GUI for device specific keyboard mappings
# Copyright (C) 2025 sezanzeb <b8x45ygc9@mozmail.com>
#
# This file is part of input-remapper.
#
# input-remapper is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# input-remapper is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with input-remapper.  If not, see <https://www.gnu.org/licenses/>.

"""Figure out where the python module needs to be placed in order to reliable import it.

Dealing with varying sys.paths is frustrating. The sys.paths in /usr are not mirrored
in /usr/local consistently. Meson uses /usr/local/python3, which ubuntus python3 does
not import from. /usr/local is ignored by python within udev. Arch does not import
from /usr/local. When in doubt, do not install into /usr/local. I do not want to deal
with ModuleNotFoundErrors. I don't care if it should be in /usr/local by convention.
I don't know how much this varies across ubuntu versions and other debian based
distributions. I want the .deb to install reliably.

sys.path samples:
endeavouros  user: ['', '/usr/lib/python313.zip', '/usr/lib/python3.13', '/usr/lib/python3.13/lib-dynload', '/usr/lib/python3.13/site-packages']
endeavouros  root: ['', '/usr/lib/python313.zip', '/usr/lib/python3.13', '/usr/lib/python3.13/lib-dynload', '/usr/lib/python3.13/site-packages']
endeavouros  udev: ['/usr/bin', '/lib/python313.zip', '/lib/python3.13', '/lib/python3.13/lib-dynload', '/lib/python3.13/site-packages']
ubuntu 25.04 user: ['', '/usr/lib/python313.zip', '/usr/lib/python3.13', '/usr/lib/python3.13/lib-dynload', '/home/user/.local/lib/python3.13/site-packages', '/usr/local/lib/python3.13/dist-packages', '/usr/lib/python3/dist-packages']
ubuntu 25.04 root: ['', '/usr/lib/python313.zip', '/usr/lib/python3.13', '/usr/lib/python3.13/lib-dynload', '/usr/local/lib/python3.13/dist-packages', '/usr/lib/python3/dist-packages']
ubuntu 25.04 udev: ['/usr/bin', '/lib/python313.zip', '/lib/python3.13', '/lib/python3.13/lib-dynload', '/lib/python3/dist-packages']
"""

import sys
import os
import subprocess
import shutil
import re
import tomllib

from install.data_files import DATA_DIR


def _key(path) -> int:
    favorability = 0

    # Good
    if re.match(r".*/lib/python3/.+-packages", path):
        # Independent of the installed python version, so it probably just continues
        # to be importable after a python version upgrade.
        favorability = 5
    elif re.match(r".*/lib/python3.+?/.+-packages", path):
        favorability = 4

    # Not meant for python code, but it probably works flawlessly
    elif re.match(r".*/lib/python3.+?/lib-dynload", path):
        favorability = 3

    # pip refuses to uninstall from standard-library directories once installed there
    elif re.match(r".*/lib/python3", path):
        favorability = 2
    elif re.match(r".*/lib/python3.+?", path):
        favorability = 1

    # Check the prefix
    # (beware, udev imports from /lib, which I think is equivalent to /usr/lib, so
    # don't require /usr as a prefix)
    if path.startswith("/usr/local"):
        # udev does not import from /usr/local
        favorability -= 5
    elif path.startswith("/home"):
        # Something like /home/user/.local/lib/python3.13/site-packages
        # The python package needs to be installed system-wide
        favorability -= 50

    if not os.path.exists(path):
        # If it doesn't exist yet, pip will apparently create it
        favorability -= 2
    elif not os.path.isdir(path):
        # If it exists but is not a dir, do not use it
        favorability -= 99

    return -favorability


def _get_packages_dir() -> str:
    """Where to install the input-remapper module to.

    For example "/usr/lib/python3.13
    """
    packages_dirs = sorted(sys.path, key=_key)
    packages_dir = packages_dirs[0]
    print(f'Picked "{packages_dir}" from {packages_dirs}')
    return packages_dir


def _get_commit_hash() -> str:
    git_call = subprocess.check_output(["git", "rev-parse", "HEAD"])
    commit = git_call.decode().strip()
    return commit


def _set_variables(target: str) -> None:
    path = os.path.join(target, "inputremapper", "installation_info.py")
    assert os.path.exists(path)

    with open(path, "r") as f:
        contents = f.read()

    with open("pyproject.toml", "rb") as f:
        version = tomllib.load(f)["project"]["version"]

    values = {
        "COMMIT_HASH": _get_commit_hash(),
        "VERSION": version,
        "DATA_DIR": f"/{DATA_DIR}",
    }

    print("Setting", values, "in", path)

    with open(path, "w") as f:
        for variable_name, value in values.items():
            contents = re.sub(
                rf"{variable_name}\s*=.+",
                f"{variable_name} = '{value}'",
                contents,
            )

        f.write(contents)


def build_input_remapper_module(root: str) -> None:
    # I'd use --prefix and --root, but
    # `pip install . --root ./build --prefix usr`
    # makes it end up in ./build/usr/local

    # if root is not /, then input-remapper is built into some other directory for the
    # purpose of merging it into / later. So regardless of root, we use the same
    # package_dir.
    package_dir = _get_packages_dir()
    if package_dir.startswith("/"):
        package_dir = package_dir[1:]

    target = os.path.join(root, package_dir)

    # Make sure we don't install input-remapper twice on that system into two different
    # paths, which commonly causes unexpected behavior. Also, we need to specifically
    # tell pip to replace an existing installation, or uninstall it beforehand.
    if root == "/":
        print("Uninstalling existing input-remapper installation")
        os.system("pip uninstall input-remapper --break-system-packages -y")
        # If root is not "/", it is probably part of a packaging script for a distro,
        # which will take care of uninstalling the old python-package for us.

    command = [
        sys.executable,
        "-m",
        "pip",
        "install",
        ".",
        "--target",
        target,
        "--no-deps",
    ]

    print("Running", " ".join(command))

    # Fix the stdout ordering in github workflows
    sys.stdout.flush()

    subprocess.check_call(command)

    # pip puts its own leftovers into ./build that we don't need.
    # This only happens, when root is set to "build".
    if "build" in root:
        if os.path.exists("./build/lib/"):
            shutil.rmtree("./build/lib/")
        if os.path.exists("./build/bdist.linux-x86_64/"):
            shutil.rmtree("./build/bdist.linux-x86_64/")

    _set_variables(target)



# -*- coding: utf-8 -*-
# input-remapper - GUI for device specific keyboard mappings
# Copyright (C) 2025 sezanzeb <b8x45ygc9@mozmail.com>
#
# This file is part of input-remapper.
#
# input-remapper is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# input-remapper is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with input-remapper.  If not, see <https://www.gnu.org/licenses/>.

"""Executes more complex patterns of keystrokes."""

from __future__ import annotations

import asyncio
from typing import List, Callable, Optional, TYPE_CHECKING

from inputremapper.ipc.shared_dict import SharedDict
from inputremapper.logging.logger import logger

if TYPE_CHECKING:
    from inputremapper.injection.macros.task import Task
    from inputremapper.injection.context import Context
    from inputremapper.configs.mapping import Mapping

InjectEventCallback = Callable[[int, int, int], None]

macro_variables = SharedDict()


class Macro:
    """Chains tasks (like `modify` or `repeat`).

    Tasks may have child_macros. Running a Macro runs Tasks, which in turn may run
    their child_macros based on certain conditions (depending on the Task).
    """

    def __init__(
        self,
        code: Optional[str],
        context: Optional[Context] = None,
        mapping: Optional[Mapping] = None,
    ):
        """Create a macro instance that can be populated with tasks.

        Parameters
        ----------
        code
            The original parsed code, for logging purposes.
        context : Context
        mapping : UIMapping
        """
        self.code = code
        self.context = context
        self.mapping = mapping

        # List of coroutines that will be called sequentially.
        # This is the compiled code
        self.tasks: List[Task] = []

        self.running = False

        self.keystroke_sleep_ms = None

    async def run(self, callback: InjectEventCallback):
        """Run the macro.

        Parameters
        ----------
        callback
            Will receive int type, code and value for an event to write
        """
        if not callable(callback):
            raise ValueError("handler is not callable")

        if self.running:
            logger.error('Tried to run already running macro "%s"', self.code)
            return

        self.keystroke_sleep_ms = self.mapping.macro_key_sleep_ms

        self.running = True

        try:
            for task in self.tasks:
                coroutine = task.run(callback)
                if asyncio.iscoroutine(coroutine):
                    await coroutine
        except Exception:
            raise
        finally:
            # done
            self.running = False

    def press_trigger(self):
        """The user pressed the trigger key down."""
        for task in self.tasks:
            task.press_trigger()

    def release_trigger(self):
        """The user released the trigger key."""
        for task in self.tasks:
            task.release_trigger()

    async def _keycode_pause(self, _=None):
        """To add a pause between keystrokes.

        This was needed at some point because it appeared that injecting keys too
        fast will prevent them from working. It probably depends on the environment.
        """
        await asyncio.sleep(self.keystroke_sleep_ms / 1000)

    def __repr__(self):
        return f'<Macro "{self.code}" at {hex(id(self))}>'

    def add_task(self, task):
        self.tasks.append(task)



# -*- coding: utf-8 -*-
# input-remapper - GUI for device specific keyboard mappings
# Copyright (C) 2025 sezanzeb <b8x45ygc9@mozmail.com>
#
# This file is part of input-remapper.
#
# input-remapper is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# input-remapper is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with input-remapper.  If not, see <https://www.gnu.org/licenses/>.

import psutil


class ProcessUtils:
    @staticmethod
    def count_python_processes(name: str) -> int:
        # This is somewhat complicated, because there might also be a "sudo <name>"
        # process.
        count = 0
        pids = psutil.pids()
        for pid in pids:
            try:
                process = psutil.Process(pid)
                cmdline = process.cmdline()
                if len(cmdline) >= 2 and "python" in cmdline[0] and name in cmdline[1]:
                    count += 1
            except Exception:
                pass

        return count



# -*- coding: utf-8 -*-
# input-remapper - GUI for device specific keyboard mappings
# Copyright (C) 2025 sezanzeb <b8x45ygc9@mozmail.com>
#
# This file is part of input-remapper.
#
# input-remapper is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# input-remapper is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with input-remapper.  If not, see <https://www.gnu.org/licenses/>.

from typing import Dict, Union, Tuple, Optional, List, Type

import evdev

import inputremapper.exceptions
import inputremapper.utils
from inputremapper.logging.logger import logger

MIN_ABS = -(2**15)  # -32768
MAX_ABS = 2**15  # 32768
DEV_NAME = "input-remapper"
DEFAULT_UINPUTS = {
    # for event codes see linux/input-event-codes.h
    "keyboard": {
        evdev.ecodes.EV_KEY: list(evdev.ecodes.KEY.keys() & evdev.ecodes.keys.keys())
    },
    "gamepad": {
        evdev.ecodes.EV_KEY: [*range(0x130, 0x13F)],  # BTN_SOUTH - BTN_THUMBR
        evdev.ecodes.EV_ABS: [
            *(
                (i, evdev.AbsInfo(0, MIN_ABS, MAX_ABS, 0, 0, 0))
                for i in range(0x00, 0x06)
            ),
            *((i, evdev.AbsInfo(0, -1, 1, 0, 0, 0)) for i in range(0x10, 0x12)),
        ],  # 6-axis and 1 hat switch
    },
    "mouse": {
        evdev.ecodes.EV_KEY: [*range(0x110, 0x118)],  # BTN_LEFT - BTN_TASK
        evdev.ecodes.EV_REL: [*range(0x00, 0x0D)],  # all REL axis
    },
}
DEFAULT_UINPUTS["keyboard + mouse"] = {
    evdev.ecodes.EV_KEY: [
        *DEFAULT_UINPUTS["keyboard"][evdev.ecodes.EV_KEY],
        *DEFAULT_UINPUTS["mouse"][evdev.ecodes.EV_KEY],
    ],
    evdev.ecodes.EV_REL: [
        *DEFAULT_UINPUTS["mouse"][evdev.ecodes.EV_REL],
    ],
}


class UInput(evdev.UInput):
    _capabilities_cache: Optional[Dict] = None

    def __init__(self, *args, **kwargs):
        name = kwargs["name"]
        logger.debug('creating UInput device: "%s"', name)
        super().__init__(*args, **kwargs)

    def can_emit(self, event: Tuple[int, int, int]):
        """Check if an event can be emitted by the UIinput.

        Wrong events might be injected if the group mappings are wrong,
        """
        # this will never change, so we cache it since evdev runs an expensive loop to
        # gather the capabilities. (can_emit is called regularly)
        if self._capabilities_cache is None:
            self._capabilities_cache = self.capabilities(absinfo=False)

        return event[1] in self._capabilities_cache.get(event[0], [])


class FrontendUInput:
    """Uinput which can not actually send events, for use in the frontend."""

    def __init__(self, *_, events=None, name="py-evdev-uinput", **__):
        # see https://python-evdev.readthedocs.io/en/latest/apidoc.html#module-evdev.uinput  # noqa pylint: disable=line-too-long
        self.events = events
        self.name = name

        logger.debug('creating fake UInput device: "%s"', self.name)

    def capabilities(self):
        return self.events


class GlobalUInputs:
    """Manages all UInputs that are shared between all injection processes."""

    def __init__(
        self,
        uinput_factory: Union[Type[UInput], Type[FrontendUInput]],
    ):
        self.devices: Dict[str, Union[UInput, FrontendUInput]] = {}
        self._uinput_factory = uinput_factory

    def __iter__(self):
        return iter(uinput for _, uinput in self.devices.items())

    @staticmethod
    def can_default_uinput_emit(target: str, type_: int, code: int) -> bool:
        """Check if the uinput with the target name is capable of the event."""
        capabilities = DEFAULT_UINPUTS.get(target, {}).get(type_)
        return capabilities is not None and code in capabilities

    @staticmethod
    def find_fitting_default_uinputs(type_: int, code: int) -> List[str]:
        """Find the names of default uinputs that are able to emit this event."""
        return [
            uinput
            for uinput in DEFAULT_UINPUTS
            if code in DEFAULT_UINPUTS[uinput].get(type_, [])
        ]

    def reset(self):
        self.devices = {}
        self.prepare_all()

    def prepare_all(self):
        """Generate UInputs."""
        for name, events in DEFAULT_UINPUTS.items():
            if name in self.devices.keys():
                continue

            self.devices[name] = self._uinput_factory(
                name=f"{DEV_NAME} {name}",
                phys=DEV_NAME,
                events=events,
            )

    def prepare_single(self, name: str):
        """Generate a single uinput.

        This has to be done in the main process before injections that use it start.
        """
        if name not in DEFAULT_UINPUTS:
            raise KeyError("Could not find a matching uinput to generate.")

        if name in self.devices:
            logger.debug('Target "%s" already exists', name)
            return

        self.devices[name] = self._uinput_factory(
            name=f"{DEV_NAME} {name}",
            phys=DEV_NAME,
            events=DEFAULT_UINPUTS[name],
        )

    def write(self, event: Tuple[int, int, int], target_uinput):
        """Write event to target uinput."""
        uinput = self.get_uinput(target_uinput)
        if not uinput:
            raise inputremapper.exceptions.UinputNotAvailable(target_uinput)

        if not uinput.can_emit(event):
            raise inputremapper.exceptions.EventNotHandled(event)

        logger.write(event, uinput)
        uinput.write(*event)
        uinput.syn()

    def get_uinput(self, name: str) -> Optional[evdev.UInput]:
        """UInput with name

        Or None if there is no uinput with this name.

        Parameters
        ----------
        name
            uniqe name of the uinput device
        """
        if name not in self.devices:
            logger.error(
                f'UInput "{name}" is unknown. '
                + f"Available: {list(self.devices.keys())}"
            )
            return None

        return self.devices.get(name)



# -*- coding: utf-8 -*-
# input-remapper - GUI for device specific keyboard mappings
# Copyright (C) 2025 sezanzeb <b8x45ygc9@mozmail.com>
#
# This file is part of input-remapper.
#
# input-remapper is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# input-remapper is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with input-remapper.  If not, see <https://www.gnu.org/licenses/>.


"""User Interface."""
from typing import Dict, Callable

from gi.repository import Gtk, GtkSource, Gdk, GObject

from inputremapper.configs.data import get_data_path
from inputremapper.configs.input_config import InputCombination
from inputremapper.configs.mapping import MappingData
from inputremapper.gui.autocompletion import Autocompletion
from inputremapper.gui.components.common import Breadcrumbs
from inputremapper.gui.components.device_groups import DeviceGroupSelection
from inputremapper.gui.components.editor import (
    MappingListBox,
    TargetSelection,
    CodeEditor,
    RecordingToggle,
    RecordingStatus,
    AutoloadSwitch,
    ReleaseCombinationSwitch,
    CombinationListbox,
    AnalogInputSwitch,
    TriggerThresholdInput,
    OutputAxisSelector,
    ReleaseTimeoutInput,
    TransformationDrawArea,
    Sliders,
    RelativeInputCutoffInput,
    KeyAxisStackSwitcher,
    RequireActiveMapping,
    GdkEventRecorder,
)
from inputremapper.gui.components.main import Stack, StatusBar
from inputremapper.gui.components.presets import PresetSelection
from inputremapper.gui.controller import Controller
from inputremapper.gui.gettext import _
from inputremapper.gui.messages.message_broker import (
    MessageBroker,
    MessageType,
)
from inputremapper.gui.messages.message_data import UserConfirmRequest
from inputremapper.gui.utils import (
    gtk_iteration,
)
from inputremapper.injection.injector import InjectorStateMessage
from inputremapper.logging.logger import logger, COMMIT_HASH, VERSION, EVDEV_VERSION

# https://cjenkins.wordpress.com/2012/05/08/use-gtksourceview-widget-in-glade/
GObject.type_register(GtkSource.View)
# GtkSource.View() also works:
# https://stackoverflow.com/questions/60126579/gtk-builder-error-quark-invalid-object-type-webkitwebview


def on_close_about(about, _):
    """Hide the about dialog without destroying it."""
    about.hide()
    return True


class UserInterface:
    """The input-remapper gtk window."""

    def __init__(
        self,
        message_broker: MessageBroker,
        controller: Controller,
    ):
        self.message_broker = message_broker
        self.controller = controller

        # all shortcuts executed when ctrl+...
        self.shortcuts: Dict[int, Callable] = {
            Gdk.KEY_q: self.controller.close,
            Gdk.KEY_r: self.controller.refresh_groups,
            Gdk.KEY_Delete: self.controller.stop_injecting,
            Gdk.KEY_n: self.controller.add_preset,
        }

        # stores the ids for all the listeners attached to the gui
        self.gtk_listeners: Dict[Callable, int] = {}

        self.message_broker.subscribe(MessageType.terminate, lambda _: self.close())

        self.builder = Gtk.Builder()
        self._build_ui()
        self.window: Gtk.Window = self.get("window")
        self.about: Gtk.Window = self.get("about-dialog")
        self.combination_editor: Gtk.Dialog = self.get("combination-editor")

        self._create_dialogs()
        self._create_components()
        self._connect_gtk_signals()
        self._connect_message_listener()

        self.window.show()
        # hide everything until stuff is populated
        self.get("vertical-wrapper").set_opacity(0)
        # if any of the next steps take a bit to complete, have the window
        # already visible (without content) to make it look more responsive.
        gtk_iteration()

        # now show the proper finished content of the window
        self.get("vertical-wrapper").set_opacity(1)

    def _build_ui(self):
        """Build the window from stylesheet and gladefile."""
        css_provider = Gtk.CssProvider()

        with open(get_data_path("style.css"), "r") as file:
            css_provider.load_from_data(bytes(file.read(), encoding="UTF-8"))

        Gtk.StyleContext.add_provider_for_screen(
            Gdk.Screen.get_default(),
            css_provider,
            Gtk.STYLE_PROVIDER_PRIORITY_APPLICATION,
        )

        gladefile = get_data_path("input-remapper.glade")
        self.builder.add_from_file(gladefile)
        self.builder.connect_signals(self)

    def _create_components(self):
        """Setup all objects which manage individual components of the ui."""
        message_broker = self.message_broker
        controller = self.controller
        DeviceGroupSelection(message_broker, controller, self.get("device_selection"))
        PresetSelection(message_broker, controller, self.get("preset_selection"))
        MappingListBox(message_broker, controller, self.get("selection_label_listbox"))
        TargetSelection(message_broker, controller, self.get("target-selector"))

        Breadcrumbs(
            message_broker,
            self.get("selected_device_name"),
            show_device_group=True,
        )
        Breadcrumbs(
            message_broker,
            self.get("selected_preset_name"),
            show_device_group=True,
            show_preset=True,
        )

        Stack(message_broker, controller, self.get("main_stack"))
        RecordingToggle(message_broker, controller, self.get("key_recording_toggle"))
        StatusBar(
            message_broker,
            controller,
            self.get("status_bar"),
            self.get("error_status_icon"),
            self.get("warning_status_icon"),
        )
        RecordingStatus(message_broker, self.get("recording_status"))
        AutoloadSwitch(message_broker, controller, self.get("preset_autoload_switch"))
        ReleaseCombinationSwitch(
            message_broker, controller, self.get("release-combination-switch")
        )
        CombinationListbox(message_broker, controller, self.get("combination-listbox"))
        AnalogInputSwitch(message_broker, controller, self.get("analog-input-switch"))
        TriggerThresholdInput(
            message_broker, controller, self.get("trigger-threshold-spin-btn")
        )
        RelativeInputCutoffInput(
            message_broker, controller, self.get("input-cutoff-spin-btn")
        )
        OutputAxisSelector(message_broker, controller, self.get("output-axis-selector"))
        KeyAxisStackSwitcher(
            message_broker,
            controller,
            self.get("editor-stack"),
            self.get("key_macro_toggle_btn"),
            self.get("analog_toggle_btn"),
        )
        ReleaseTimeoutInput(
            message_broker, controller, self.get("release-timeout-spin-button")
        )
        TransformationDrawArea(
            message_broker, controller, self.get("transformation-draw-area")
        )
        Sliders(
            message_broker,
            controller,
            self.get("gain-scale"),
            self.get("deadzone-scale"),
            self.get("expo-scale"),
        )

        GdkEventRecorder(self.window, self.get("gdk-event-recorder-label"))

        RequireActiveMapping(
            message_broker,
            self.get("edit-combination-btn"),
            require_recorded_input=True,
        )
        RequireActiveMapping(
            message_broker,
            self.get("output"),
            require_recorded_input=True,
        )
        RequireActiveMapping(
            message_broker,
            self.get("delete-mapping"),
            require_recorded_input=False,
        )

        # code editor and autocompletion
        code_editor = CodeEditor(message_broker, controller, self.get("code_editor"))
        autocompletion = Autocompletion(message_broker, controller, code_editor)
        autocompletion.set_relative_to(self.get("code_editor_container"))
        self.autocompletion = autocompletion  # only for testing

    def _create_dialogs(self):
        """Setup different dialogs, such as the about page."""
        self.about.connect("delete-event", on_close_about)
        # set_position needs to be done once initially, otherwise the
        # dialog is not centered when it is opened for the first time
        self.about.set_position(Gtk.WindowPosition.CENTER_ON_PARENT)
        self.get("version-label").set_text(
            f"input-remapper {VERSION} {COMMIT_HASH[:7]}"
            f"\npython-evdev {EVDEV_VERSION}"
            if EVDEV_VERSION
            else ""
        )

    def _connect_gtk_signals(self):
        self.get("delete_preset").connect(
            "clicked", lambda *_: self.controller.delete_preset()
        )
        self.get("copy_preset").connect(
            "clicked", lambda *_: self.controller.copy_preset()
        )
        self.get("create_preset").connect(
            "clicked", lambda *_: self.controller.add_preset()
        )
        self.get("apply_preset").connect(
            "clicked", lambda *_: self.controller.start_injecting()
        )
        self.get("stop_injection_preset_page").connect(
            "clicked", lambda *_: self.controller.stop_injecting()
        )
        self.get("stop_injection_editor_page").connect(
            "clicked", lambda *_: self.controller.stop_injecting()
        )
        self.get("rename-button").connect("clicked", self.on_gtk_rename_clicked)
        self.get("preset_name_input").connect(
            "key-release-event", self.on_gtk_preset_name_input_return
        )
        self.get("create_mapping_button").connect(
            "clicked", lambda *_: self.controller.create_mapping()
        )
        self.get("delete-mapping").connect(
            "clicked", lambda *_: self.controller.delete_mapping()
        )
        self.combination_editor.connect(
            # it only takes self as argument, but delete-events provides more
            # probably a gtk bug
            "delete-event",
            lambda dialog, *_: Gtk.Widget.hide_on_delete(dialog),
        )
        self.get("edit-combination-btn").connect(
            "clicked", lambda *_: self.combination_editor.show()
        )
        self.get("remove-event-btn").connect(
            "clicked", lambda *_: self.controller.remove_event()
        )
        self.connect_shortcuts()

    def _connect_message_listener(self):
        self.message_broker.subscribe(
            MessageType.mapping, self.update_combination_label
        )
        self.message_broker.subscribe(
            MessageType.injector_state, self.on_injector_state_msg
        )
        self.message_broker.subscribe(
            MessageType.user_confirm_request, self._on_user_confirm_request
        )

    def _create_dialog(self, primary: str, secondary: str) -> Gtk.MessageDialog:
        """Create a message dialog with cancel and confirm buttons."""
        message_dialog = Gtk.MessageDialog(
            self.window,
            Gtk.DialogFlags.MODAL | Gtk.DialogFlags.DESTROY_WITH_PARENT,
            Gtk.MessageType.QUESTION,
            Gtk.ButtonsType.NONE,
            primary,
        )

        if secondary:
            message_dialog.format_secondary_text(secondary)

        message_dialog.add_button("Cancel", Gtk.ResponseType.CANCEL)

        confirm_button = message_dialog.add_button("Confirm", Gtk.ResponseType.ACCEPT)
        confirm_button.get_style_context().add_class(Gtk.STYLE_CLASS_DESTRUCTIVE_ACTION)

        return message_dialog

    def _on_user_confirm_request(self, msg: UserConfirmRequest):
        # if the message contains a line-break, use the first chunk for the primary
        # message, and the rest for the secondary message.
        chunks = msg.msg.split("\n")
        primary = chunks[0]
        secondary = " ".join(chunks[1:])

        message_dialog = self._create_dialog(primary, secondary)

        response = message_dialog.run()
        msg.respond(response == Gtk.ResponseType.ACCEPT)

        message_dialog.hide()

    def on_injector_state_msg(self, msg: InjectorStateMessage):
        """Update the ui to reflect the status of the injector."""
        stop_injection_preset_page: Gtk.Button = self.get("stop_injection_preset_page")
        stop_injection_editor_page: Gtk.Button = self.get("stop_injection_editor_page")
        recording_toggle: Gtk.ToggleButton = self.get("key_recording_toggle")

        if msg.active():
            stop_injection_preset_page.set_opacity(1)
            stop_injection_editor_page.set_opacity(1)
            stop_injection_preset_page.set_sensitive(True)
            stop_injection_editor_page.set_sensitive(True)
            recording_toggle.set_opacity(0.5)
        else:
            stop_injection_preset_page.set_opacity(0.5)
            stop_injection_editor_page.set_opacity(0.5)
            stop_injection_preset_page.set_sensitive(True)
            stop_injection_editor_page.set_sensitive(True)
            recording_toggle.set_opacity(1)

    def disconnect_shortcuts(self):
        """Stop listening for shortcuts.

        e.g. when recording key combinations
        """
        try:
            self.window.disconnect(self.gtk_listeners.pop(self.on_gtk_shortcut))
        except KeyError:
            logger.debug("key listeners seem to be not connected")

    def connect_shortcuts(self):
        """Start listening for shortcuts."""
        if not self.gtk_listeners.get(self.on_gtk_shortcut):
            self.gtk_listeners[self.on_gtk_shortcut] = self.window.connect(
                "key-press-event", self.on_gtk_shortcut
            )

    def get(self, name: str):
        """Get a widget from the window."""
        return self.builder.get_object(name)

    def close(self):
        """Close the window."""
        logger.debug("Closing window")
        self.window.hide()

    def update_combination_label(self, mapping: MappingData):
        """Listens for mapping and updates the combination label."""
        label: Gtk.Label = self.get("combination-label")
        if mapping.input_combination.beautify() == label.get_label():
            return
        if mapping.input_combination == InputCombination.empty_combination():
            label.set_opacity(0.5)
            label.set_label(_("no input configured"))
            return

        label.set_opacity(1)
        label.set_label(mapping.input_combination.beautify())

    def on_gtk_shortcut(self, _, event: Gdk.EventKey):
        """Execute shortcuts."""
        if event.state & Gdk.ModifierType.CONTROL_MASK:
            try:
                self.shortcuts[event.keyval]()
            except KeyError:
                pass

    def on_gtk_close(self, *_):
        self.controller.close()

    def on_gtk_about_clicked(self, _):
        """Show the about/help dialog."""
        self.about.show()

    def on_gtk_about_key_press(self, _, event):
        """Hide the about/help dialog."""
        gdk_keycode = event.get_keyval()[1]
        if gdk_keycode == Gdk.KEY_Escape:
            self.about.hide()

    def on_gtk_rename_clicked(self, *_):
        name = self.get("preset_name_input").get_text()
        self.controller.rename_preset(name)
        self.get("preset_name_input").set_text("")

    def on_gtk_preset_name_input_return(self, _, event: Gdk.EventKey):
        if event.keyval == Gdk.KEY_Return:
            self.on_gtk_rename_clicked()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# input-remapper - GUI for device specific keyboard mappings
# Copyright (C) 2025 sezanzeb <b8x45ygc9@mozmail.com>
#
# This file is part of input-remapper.
#
# input-remapper is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# input-remapper is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with input-remapper.  If not, see <https://www.gnu.org/licenses/>.

import time
import unittest

import gi


gi.require_version("Gdk", "3.0")
gi.require_version("Gtk", "3.0")
gi.require_version("GtkSource", "4")
gi.require_version("GLib", "2.0")

from inputremapper.gui.utils import gtk_iteration, debounce, debounce_manager

from tests.lib.test_setup import test_setup


@test_setup
class TestDebounce(unittest.TestCase):
    def test_debounce(self):
        calls = 0

        class A:
            @debounce(20)
            def foo(self):
                nonlocal calls
                calls += 1

        # two methods with the same name don't confuse debounce
        class B:
            @debounce(20)
            def foo(self):
                nonlocal calls
                calls += 1

        a = A()
        b = B()

        self.assertEqual(calls, 0)

        a.foo()
        gtk_iteration()
        self.assertEqual(calls, 0)

        b.foo()
        gtk_iteration()
        self.assertEqual(calls, 0)

        time.sleep(0.021)
        gtk_iteration()
        self.assertEqual(calls, 2)

        a.foo()
        b.foo()
        a.foo()
        b.foo()
        gtk_iteration()
        self.assertEqual(calls, 2)

        time.sleep(0.021)
        gtk_iteration()
        self.assertEqual(calls, 4)

    def test_run_all_now(self):
        calls = 0

        class A:
            @debounce(20)
            def foo(self):
                nonlocal calls
                calls += 1

        a = A()
        a.foo()
        gtk_iteration()
        self.assertEqual(calls, 0)

        debounce_manager.run_all_now()
        self.assertEqual(calls, 1)

        # waiting for some time will not call it again
        time.sleep(0.021)
        gtk_iteration()
        self.assertEqual(calls, 1)

    def test_stop_all(self):
        calls = 0

        class A:
            @debounce(20)
            def foo(self):
                nonlocal calls
                calls += 1

        a = A()
        a.foo()
        gtk_iteration()
        self.assertEqual(calls, 0)

        debounce_manager.stop_all()

        # waiting for some time will not call it
        time.sleep(0.021)
        gtk_iteration()
        self.assertEqual(calls, 0)

    def test_stop(self):
        calls = 0

        class A:
            @debounce(20)
            def foo(self):
                nonlocal calls
                calls += 1

        a = A()
        a.foo()
        gtk_iteration()
        self.assertEqual(calls, 0)

        debounce_manager.stop(a, a.foo)

        # waiting for some time will not call it
        time.sleep(0.021)
        gtk_iteration()
        self.assertEqual(calls, 0)


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# input-remapper - GUI for device specific keyboard mappings
# Copyright (C) 2025 sezanzeb <b8x45ygc9@mozmail.com>
#
# This file is part of input-remapper.
#
# input-remapper is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# input-remapper is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with input-remapper.  If not, see <https://www.gnu.org/licenses/>.


import unittest

from inputremapper.injection.numlock import is_numlock_on, set_numlock, ensure_numlock
from tests.lib.test_setup import test_setup


@test_setup
class TestNumlock(unittest.TestCase):
    def test_numlock(self):
        before = is_numlock_on()

        set_numlock(not before)  # should change
        self.assertEqual(not before, is_numlock_on())

        @ensure_numlock
        def wrapped_1():
            set_numlock(not is_numlock_on())

        @ensure_numlock
        def wrapped_2():
            pass

        # should not change
        wrapped_1()
        self.assertEqual(not before, is_numlock_on())
        wrapped_2()
        self.assertEqual(not before, is_numlock_on())

        # toggle one more time to restore the previous configuration
        set_numlock(before)
        self.assertEqual(before, is_numlock_on())


if __name__ == "__main__":
    unittest.main()



import unittest
from unittest.mock import MagicMock

import gi
from evdev.ecodes import EV_KEY, KEY_A

gi.require_version("Gdk", "3.0")
gi.require_version("Gtk", "3.0")
gi.require_version("GLib", "2.0")
gi.require_version("GtkSource", "4")
from gi.repository import Gtk, Gdk, GLib

from inputremapper.gui.utils import gtk_iteration
from inputremapper.gui.messages.message_broker import MessageBroker, MessageType
from inputremapper.gui.user_interface import UserInterface
from inputremapper.configs.mapping import MappingData
from inputremapper.configs.input_config import InputCombination, InputConfig
from tests.lib.test_setup import test_setup


@test_setup
class TestUserInterface(unittest.TestCase):
    def setUp(self) -> None:
        self.message_broker = MessageBroker()
        self.controller_mock = MagicMock()
        self.user_interface = UserInterface(self.message_broker, self.controller_mock)

    def tearDown(self) -> None:
        super().tearDown()
        self.message_broker.signal(MessageType.terminate)
        GLib.timeout_add(0, self.user_interface.window.destroy)
        GLib.timeout_add(0, Gtk.main_quit)
        Gtk.main()

    def test_shortcut(self):
        mock = MagicMock()
        self.user_interface.shortcuts[Gdk.KEY_x] = mock

        event = Gdk.Event()
        event.key.keyval = Gdk.KEY_x
        event.key.state = Gdk.ModifierType.SHIFT_MASK
        self.user_interface.window.emit("key-press-event", event)
        gtk_iteration()
        mock.assert_not_called()

        event.key.state = Gdk.ModifierType.CONTROL_MASK
        self.user_interface.window.emit("key-press-event", event)
        gtk_iteration()
        mock.assert_called_once()

        mock.reset_mock()
        event.key.keyval = Gdk.KEY_y
        self.user_interface.window.emit("key-press-event", event)
        gtk_iteration()
        mock.assert_not_called()

    def test_connected_shortcuts(self):
        should_be_connected = {Gdk.KEY_q, Gdk.KEY_r, Gdk.KEY_Delete, Gdk.KEY_n}
        connected = set(self.user_interface.shortcuts.keys())
        self.assertEqual(connected, should_be_connected)

        self.assertIs(
            self.user_interface.shortcuts[Gdk.KEY_q], self.controller_mock.close
        )
        self.assertIs(
            self.user_interface.shortcuts[Gdk.KEY_r],
            self.controller_mock.refresh_groups,
        )
        self.assertIs(
            self.user_interface.shortcuts[Gdk.KEY_Delete],
            self.controller_mock.stop_injecting,
        )

    def test_connect_disconnect_shortcuts(self):
        mock = MagicMock()
        self.user_interface.shortcuts[Gdk.KEY_x] = mock

        event = Gdk.Event()
        event.key.keyval = Gdk.KEY_x
        event.key.state = Gdk.ModifierType.CONTROL_MASK
        self.user_interface.disconnect_shortcuts()
        self.user_interface.window.emit("key-press-event", event)
        gtk_iteration()
        mock.assert_not_called()

        self.user_interface.connect_shortcuts()
        gtk_iteration()
        self.user_interface.window.emit("key-press-event", event)
        gtk_iteration()
        mock.assert_called_once()

    def test_combination_label_shows_combination(self):
        self.message_broker.publish(
            MappingData(
                input_combination=InputCombination(
                    [InputConfig(type=EV_KEY, code=KEY_A)]
                ),
                name="foo",
            )
        )
        gtk_iteration()
        label: Gtk.Label = self.user_interface.get("combination-label")
        self.assertEqual(label.get_text(), "a")
        self.assertEqual(label.get_opacity(), 1)

    def test_combination_label_shows_text_when_empty_mapping(self):
        self.message_broker.publish(MappingData())
        gtk_iteration()
        label: Gtk.Label = self.user_interface.get("combination-label")
        self.assertEqual(label.get_text(), "no input configured")

        # 0.5 != 0.501960..., for whatever reason this number is all screwed up
        self.assertAlmostEqual(label.get_opacity(), 0.5, delta=0.1)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# input-remapper - GUI for device specific keyboard mappings
# Copyright (C) 2025 sezanzeb <b8x45ygc9@mozmail.com>
#
# This file is part of input-remapper.
#
# input-remapper is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# input-remapper is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with input-remapper.  If not, see <https://www.gnu.org/licenses/>.

from __future__ import annotations

import asyncio
import copy
import os
import shutil
import time
from pickle import UnpicklingError

import psutil

from tests.lib.constants import EVENT_READ_TIMEOUT
from tests.lib.fixtures import fixtures
from tests.lib.logger import logger
from tests.lib.patches import uinputs
from tests.lib.pipes import (
    uinput_write_history_pipe,
    uinput_write_history,
    pending_events,
    setup_pipe,
)
from tests.lib.tmp import tmp

# TODO on it. You don't need a framework for this by the way:
# don't import anything from input_remapper gloablly here, because some files execute
# code when imported, which can screw up patches. I wish we had a dependency injection
# framework that patches together the dependencies during runtime...


environ_copy = copy.deepcopy(os.environ)


def join_children():
    """Wait for child processes to exit. Stop them if it takes too long."""
    this = psutil.Process(os.getpid())

    i = 0
    time.sleep(EVENT_READ_TIMEOUT)
    children = this.children(recursive=True)
    while len([c for c in children if c.status() != "zombie"]) > 0:
        for child in children:
            if i > 10:
                child.kill()
                logger.info("Killed pid %s because it didn't finish in time", child.pid)

        children = this.children(recursive=True)
        time.sleep(EVENT_READ_TIMEOUT)
        i += 1


def clear_write_history():
    """Empty the history in preparation for the next test."""
    while len(uinput_write_history) > 0:
        uinput_write_history.pop()
    while uinput_write_history_pipe[0].poll():
        uinput_write_history_pipe[0].recv()


def quick_cleanup(log=True):
    """Reset the applications state."""
    # TODO no:
    # Reminder: before patches are applied in test.py, no inputremapper module
    # may be imported. So tests.lib imports them just-in-time in functions instead.
    from inputremapper.injection.macros.macro import macro_variables
    from inputremapper.configs.keyboard_layout import keyboard_layout
    from inputremapper.gui.utils import debounce_manager

    if log:
        logger.info("Quick cleanup...")

    debounce_manager.stop_all()

    for device in list(pending_events.keys()):
        try:
            while pending_events[device][1].poll():
                pending_events[device][1].recv()
        except (UnpicklingError, EOFError):
            pass

        # setup new pipes for the next test
        pending_events[device][1].close()
        pending_events[device][0].close()
        del pending_events[device]
        setup_pipe(device)

    try:
        if asyncio.get_event_loop().is_running():
            for task in asyncio.all_tasks():
                task.cancel()
    except RuntimeError:
        # happens when the event loop disappears for magical reasons
        # create a fresh event loop
        asyncio.set_event_loop(asyncio.new_event_loop())

    if macro_variables.process is not None and not macro_variables.process.is_alive():
        # nothing should stop the process during runtime, if it has been started by
        # the injector once
        raise AssertionError("the SharedDict manager is not running anymore")

    if macro_variables.process is not None:
        macro_variables._stop()

    join_children()

    macro_variables.start()

    if os.path.exists(tmp):
        shutil.rmtree(tmp)

    keyboard_layout.populate()

    clear_write_history()

    for name in list(uinputs.keys()):
        del uinputs[name]

    # for device in list(active_macros.keys()):
    #    del active_macros[device]
    # for device in list(unreleased.keys()):
    #    del unreleased[device]
    fixtures.reset()
    os.environ.update(environ_copy)
    for device in list(os.environ.keys()):
        if device not in environ_copy:
            del os.environ[device]

    for _, pipe in pending_events.values():
        assert not pipe.poll()

    assert macro_variables.is_alive(1)

    if log:
        logger.info("Quick cleanup done")


def cleanup():
    """Reset the applications state.

    Using this is slower, usually quick_cleanup() is sufficient.
    """
    from inputremapper.groups import groups

    logger.info("Cleanup...")

    os.system("pkill -f input-remapper-service")
    os.system("pkill -f input-remapper-control")
    time.sleep(0.05)

    quick_cleanup(log=False)
    groups.refresh()

    logger.info("Cleanup done")



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# input-remapper - GUI for device specific keyboard mappings
# Copyright (C) 2025 sezanzeb <b8x45ygc9@mozmail.com>
#
# This file is part of input-remapper.
#
# input-remapper is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# input-remapper is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with input-remapper.  If not, see <https://www.gnu.org/licenses/>.


import unittest

from evdev.ecodes import (
    EV_KEY,
    EV_ABS,
    EV_REL,
    BTN_C,
    BTN_B,
    BTN_A,
    BTN_LEFT,
    BTN_MIDDLE,
    BTN_RIGHT,
    BTN_SIDE,
    BTN_EXTRA,
    BTN_FORWARD,
    BTN_BACK,
    BTN_TASK,
    REL_X,
    REL_Y,
    REL_WHEEL,
    REL_HWHEEL,
    ABS_RY,
    ABS_X,
    ABS_HAT0Y,
    ABS_HAT0X,
    KEY_A,
    KEY_LEFTSHIFT,
    KEY_RIGHTALT,
    KEY_LEFTCTRL,
)

from inputremapper.configs.input_config import InputCombination, InputConfig
from tests.lib.test_setup import test_setup


@test_setup
class TestInputConfig(unittest.TestCase):
    def test_input_config(self):
        test_cases = [
            # basic test, nothing fancy here
            {
                "input": {
                    "type": EV_KEY,
                    "code": KEY_A,
                    "origin_hash": "foo",
                },
                "properties": {
                    "type": EV_KEY,
                    "code": KEY_A,
                    "origin_hash": "foo",
                    "input_match_hash": (EV_KEY, KEY_A, "foo"),
                    "defines_analog_input": False,
                    "type_and_code": (EV_KEY, KEY_A),
                },
                "methods": [
                    {
                        "name": "description",
                        "args": (),
                        "kwargs": {},
                        "return": "a",
                    },
                    {
                        "name": "__hash__",
                        "args": (),
                        "kwargs": {},
                        "return": hash((EV_KEY, KEY_A, "foo", None)),
                    },
                ],
            },
            # removes analog_threshold
            {
                "input": {
                    "type": EV_KEY,
                    "code": KEY_A,
                    "origin_hash": "foo",
                    "analog_threshold": 10,
                },
                "properties": {
                    "type": EV_KEY,
                    "code": KEY_A,
                    "origin_hash": "foo",
                    "analog_threshold": None,
                    "input_match_hash": (EV_KEY, KEY_A, "foo"),
                    "defines_analog_input": False,
                    "type_and_code": (EV_KEY, KEY_A),
                },
                "methods": [
                    {
                        "name": "description",
                        "args": (),
                        "kwargs": {},
                        "return": "a",
                    },
                    {
                        "name": "__hash__",
                        "args": (),
                        "kwargs": {},
                        "return": hash((EV_KEY, KEY_A, "foo", None)),
                    },
                ],
            },
            # abs to btn
            {
                "input": {
                    "type": EV_ABS,
                    "code": ABS_X,
                    "origin_hash": "foo",
                    "analog_threshold": 10,
                },
                "properties": {
                    "type": EV_ABS,
                    "code": ABS_X,
                    "origin_hash": "foo",
                    "analog_threshold": 10,
                    "input_match_hash": (EV_ABS, ABS_X, "foo"),
                    "defines_analog_input": False,
                    "type_and_code": (EV_ABS, ABS_X),
                },
                "methods": [
                    {
                        "name": "description",
                        "args": (),
                        "kwargs": {},
                        "return": "Joystick-X Right 10%",
                    },
                    {
                        "name": "description",
                        "args": (),
                        "kwargs": {"exclude_threshold": True},
                        "return": "Joystick-X Right",
                    },
                    {
                        "name": "description",
                        "args": (),
                        "kwargs": {
                            "exclude_threshold": True,
                            "exclude_direction": True,
                        },
                        "return": "Joystick-X",
                    },
                    {
                        "name": "__hash__",
                        "args": (),
                        "kwargs": {},
                        "return": hash((EV_ABS, ABS_X, "foo", 10)),
                    },
                ],
            },
            # abs to btn with d-pad
            {
                "input": {
                    "type": EV_ABS,
                    "code": ABS_HAT0Y,
                    "origin_hash": "foo",
                    "analog_threshold": 10,
                },
                "properties": {
                    "type": EV_ABS,
                    "code": ABS_HAT0Y,
                    "origin_hash": "foo",
                    "analog_threshold": 10,
                    "input_match_hash": (EV_ABS, ABS_HAT0Y, "foo"),
                    "defines_analog_input": False,
                    "type_and_code": (EV_ABS, ABS_HAT0Y),
                },
                "methods": [
                    {
                        "name": "description",
                        "args": (),
                        "kwargs": {},
                        "return": "DPad-Y Down 10%",
                    },
                    {
                        "name": "__hash__",
                        "args": (),
                        "kwargs": {},
                        "return": hash((EV_ABS, ABS_HAT0Y, "foo", 10)),
                    },
                ],
            },
            # rel to btn
            {
                "input": {
                    "type": EV_REL,
                    "code": REL_Y,
                    "origin_hash": "foo",
                    "analog_threshold": 10,
                },
                "properties": {
                    "type": EV_REL,
                    "code": REL_Y,
                    "origin_hash": "foo",
                    "analog_threshold": 10,
                    "input_match_hash": (EV_REL, REL_Y, "foo"),
                    "defines_analog_input": False,
                    "type_and_code": (EV_REL, REL_Y),
                },
                "methods": [
                    {
                        "name": "description",
                        "args": (),
                        "kwargs": {},
                        "return": "Y Down 10",
                    },
                    {
                        "name": "__hash__",
                        "args": (),
                        "kwargs": {},
                        "return": hash((EV_REL, REL_Y, "foo", 10)),
                    },
                ],
            },
            # abs as axis
            {
                "input": {
                    "type": EV_ABS,
                    "code": ABS_X,
                    "origin_hash": "foo",
                    "analog_threshold": 0,
                },
                "properties": {
                    "type": EV_ABS,
                    "code": ABS_X,
                    "origin_hash": "foo",
                    "analog_threshold": None,
                    "input_match_hash": (EV_ABS, ABS_X, "foo"),
                    "defines_analog_input": True,
                    "type_and_code": (EV_ABS, ABS_X),
                },
                "methods": [
                    {
                        "name": "description",
                        "args": (),
                        "kwargs": {},
                        "return": "Joystick-X",
                    },
                    {
                        "name": "description",
                        "args": (),
                        "kwargs": {
                            "exclude_threshold": True,
                            "exclude_direction": True,
                        },
                        "return": "Joystick-X",
                    },
                    {
                        "name": "__hash__",
                        "args": (),
                        "kwargs": {},
                        "return": hash((EV_ABS, ABS_X, "foo", None)),
                    },
                ],
            },
            # rel as axis
            {
                "input": {
                    "type": EV_REL,
                    "code": REL_WHEEL,
                    "origin_hash": "foo",
                },
                "properties": {
                    "type": EV_REL,
                    "code": REL_WHEEL,
                    "origin_hash": "foo",
                    "analog_threshold": None,
                    "input_match_hash": (EV_REL, REL_WHEEL, "foo"),
                    "defines_analog_input": True,
                    "type_and_code": (EV_REL, REL_WHEEL),
                },
                "methods": [
                    {
                        "name": "description",
                        "args": (),
                        "kwargs": {},
                        "return": "Wheel",
                    },
                    {
                        "name": "__hash__",
                        "args": (),
                        "kwargs": {},
                        "return": hash((EV_REL, REL_WHEEL, "foo", None)),
                    },
                ],
            },
        ]
        for test_case in test_cases:
            input_config = InputConfig(**test_case["input"])
            for property_, value in test_case["properties"].items():
                self.assertEqual(
                    value,
                    getattr(input_config, property_),
                    f"property mismatch for input: {test_case['input']} "
                    f"property: {property_} expected value: {value}",
                )
            for method in test_case["methods"]:
                self.assertEqual(
                    method["return"],
                    getattr(input_config, method["name"])(
                        *method["args"], **method["kwargs"]
                    ),
                    f"wrong method return for input: {test_case['input']} "
                    f"method: {method}",
                )

    def test_is_immutable(self):
        input_config = InputConfig(type=1, code=2)
        with self.assertRaises(TypeError):
            input_config.origin_hash = "foo"


@test_setup
class TestInputCombination(unittest.TestCase):
    def test_eq(self):
        a = InputCombination(
            [
                InputConfig(type=EV_REL, code=REL_X, value=1, origin_hash="1234"),
                InputConfig(type=EV_KEY, code=KEY_A, value=1, origin_hash="abcd"),
            ]
        )
        b = InputCombination(
            [
                InputConfig(type=EV_REL, code=REL_X, value=1, origin_hash="1234"),
                InputConfig(type=EV_KEY, code=KEY_A, value=1, origin_hash="abcd"),
            ]
        )
        self.assertEqual(a, b)

    def test_not_eq(self):
        a = InputCombination(
            [
                InputConfig(type=EV_REL, code=REL_X, value=1, origin_hash="2345"),
                InputConfig(type=EV_KEY, code=KEY_A, value=1, origin_hash="bcde"),
            ]
        )
        b = InputCombination(
            [
                InputConfig(type=EV_REL, code=REL_X, value=1, origin_hash="1234"),
                InputConfig(type=EV_KEY, code=KEY_A, value=1, origin_hash="abcd"),
            ]
        )
        self.assertNotEqual(a, b)

    def test_can_be_used_as_dict_key(self):
        dict_ = {
            InputCombination(
                [
                    InputConfig(type=EV_REL, code=REL_X, value=1, origin_hash="1234"),
                    InputConfig(type=EV_KEY, code=KEY_A, value=1, origin_hash="abcd"),
                ]
            ): "foo"
        }
        key = InputCombination(
            [
                InputConfig(type=EV_REL, code=REL_X, value=1, origin_hash="1234"),
                InputConfig(type=EV_KEY, code=KEY_A, value=1, origin_hash="abcd"),
            ]
        )
        self.assertEqual(dict_.get(key), "foo")

    def test_get_permutations(self):
        key_1 = InputCombination(InputCombination.from_tuples((1, 3, 1)))
        self.assertEqual(len(key_1.get_permutations()), 1)
        self.assertEqual(key_1.get_permutations()[0], key_1)

        key_2 = InputCombination(InputCombination.from_tuples((1, 3, 1), (1, 5, 1)))
        self.assertEqual(len(key_2.get_permutations()), 1)
        self.assertEqual(key_2.get_permutations()[0], key_2)

        key_3 = InputCombination(
            InputCombination.from_tuples((1, 3, 1), (1, 5, 1), (1, 7, 1))
        )
        self.assertEqual(len(key_3.get_permutations()), 2)
        self.assertEqual(
            key_3.get_permutations()[0],
            InputCombination(
                InputCombination.from_tuples((1, 3, 1), (1, 5, 1), (1, 7, 1))
            ),
        )
        self.assertEqual(
            key_3.get_permutations()[1],
            InputCombination(
                InputCombination.from_tuples((1, 5, 1), (1, 3, 1), (1, 7, 1))
            ),
        )

    def test_is_problematic(self):
        key_1 = InputCombination(
            InputCombination.from_tuples((1, KEY_LEFTSHIFT, 1), (1, 5, 1))
        )
        self.assertTrue(key_1.is_problematic())

        key_2 = InputCombination(
            InputCombination.from_tuples((1, KEY_RIGHTALT, 1), (1, 5, 1))
        )
        self.assertTrue(key_2.is_problematic())

        key_3 = InputCombination(
            InputCombination.from_tuples((1, 3, 1), (1, KEY_LEFTCTRL, 1))
        )
        self.assertTrue(key_3.is_problematic())

        key_4 = InputCombination(InputCombination.from_tuples((1, 3, 1)))
        self.assertFalse(key_4.is_problematic())

        key_5 = InputCombination(InputCombination.from_tuples((1, 3, 1), (1, 5, 1)))
        self.assertFalse(key_5.is_problematic())

    def test_init(self):
        self.assertRaises(TypeError, lambda: InputCombination(1))
        self.assertRaises(TypeError, lambda: InputCombination(None))
        self.assertRaises(TypeError, lambda: InputCombination([1]))
        self.assertRaises(TypeError, lambda: InputCombination((1,)))
        self.assertRaises(TypeError, lambda: InputCombination((1, 2)))
        self.assertRaises(TypeError, lambda: InputCombination("1"))
        self.assertRaises(TypeError, lambda: InputCombination("(1,2,3)"))
        self.assertRaises(
            TypeError,
            lambda: InputCombination(((1, 2, 3), (1, 2, 3), None)),
        )

        # those don't raise errors
        InputCombination(({"type": 1, "code": 2}, {"type": 1, "code": 1}))
        InputCombination(({"type": 1, "code": 2},))
        InputCombination(({"type": "1", "code": "2"},))
        InputCombination([InputConfig(type=1, code=2, analog_threshold=3)])
        InputCombination(
            (
                {"type": 1, "code": 2},
                {"type": "1", "code": "2"},
                InputConfig(type=1, code=2),
            )
        )

    def test_to_config(self):
        c1 = InputCombination([InputConfig(type=1, code=2, analog_threshold=3)])
        c2 = InputCombination(
            (
                InputConfig(type=1, code=2, analog_threshold=3),
                InputConfig(type=4, code=5, analog_threshold=6),
            )
        )
        # analog_threshold is removed for key events
        self.assertEqual(c1.to_config(), ({"type": 1, "code": 2},))
        self.assertEqual(
            c2.to_config(),
            ({"type": 1, "code": 2}, {"type": 4, "code": 5, "analog_threshold": 6}),
        )

    def test_beautify(self):
        # not an integration test, but I have all the selection_label tests here already
        self.assert_beautify_single(EV_KEY, KEY_A, 1, "a")
        self.assert_beautify_single(EV_KEY, KEY_A, 1, "a")
        self.assert_beautify_single(EV_ABS, ABS_HAT0Y, -1, "DPad-Y Up")
        self.assert_beautify_single(EV_KEY, BTN_A, 1, "Button A")
        self.assert_beautify_single(EV_KEY, 1234, 1, "unknown (1, 1234)")
        self.assert_beautify_single(EV_ABS, ABS_HAT0X, -1, "DPad-X Left")
        self.assert_beautify_single(EV_ABS, ABS_HAT0Y, -1, "DPad-Y Up")
        self.assert_beautify_single(EV_KEY, BTN_A, 1, "Button A")
        self.assert_beautify_single(EV_ABS, ABS_X, 1, "Joystick-X Right")
        self.assert_beautify_single(EV_ABS, ABS_RY, 1, "Joystick-RY Down")
        self.assert_beautify_single(EV_REL, REL_HWHEEL, 1, "Wheel Right")
        self.assert_beautify_single(EV_REL, REL_WHEEL, -1, "Wheel Down")

        # region "mouse buttons"
        self.assert_beautify_single(EV_KEY, BTN_LEFT, 1, "Mouse Button LEFT")
        self.assert_beautify_single(EV_KEY, BTN_MIDDLE, 1, "Mouse Button MIDDLE")
        self.assert_beautify_single(EV_KEY, BTN_RIGHT, 1, "Mouse Button RIGHT")
        self.assert_beautify_single(EV_KEY, BTN_SIDE, 1, "Mouse Button 4")
        self.assert_beautify_single(EV_KEY, BTN_EXTRA, 1, "Mouse Button 5")
        self.assert_beautify_single(EV_KEY, BTN_FORWARD, 1, "Mouse Button 6")
        self.assert_beautify_single(EV_KEY, BTN_BACK, 1, "Mouse Button 7")
        self.assert_beautify_single(EV_KEY, BTN_TASK, 1, "Mouse Button 8")

        # Mouse buttons 9+ do not have an evdev name, and so must be expressed
        # as an offset from BTN_LEFT AKA "Button 1". Subtract 1 for base index
        # so that `btn_mouse_base + 1` would be "Button 1".
        btn_mouse_base: int = BTN_LEFT - 1

        self.assert_beautify_single(EV_KEY, btn_mouse_base + 9, 1, "Mouse Button 9")
        self.assert_beautify_single(EV_KEY, btn_mouse_base + 10, 1, "Mouse Button 10")
        # endregion "mouse buttons"

        # combinations
        self.assertEqual(
            InputCombination(
                InputCombination.from_tuples(
                    (EV_KEY, BTN_A, 1),
                    (EV_KEY, BTN_B, 1),
                    (EV_KEY, BTN_C, 1),
                ),
            ).beautify(),
            "Button A + Button B + Button C",
        )

    def test_find_analog_input_config(self):
        analog_input = InputConfig(type=EV_REL, code=REL_X)

        combination = InputCombination(
            (
                InputConfig(type=EV_KEY, code=BTN_MIDDLE),
                InputConfig(type=EV_REL, code=REL_Y, analog_threshold=1),
                analog_input,
            )
        )
        self.assertIsNone(combination.find_analog_input_config(type_=EV_ABS))
        self.assertEqual(
            combination.find_analog_input_config(type_=EV_REL), analog_input
        )
        self.assertEqual(combination.find_analog_input_config(), analog_input)

        combination = InputCombination(
            (
                InputConfig(type=EV_REL, code=REL_X, analog_threshold=1),
                InputConfig(type=EV_KEY, code=BTN_MIDDLE),
            )
        )
        self.assertIsNone(combination.find_analog_input_config(type_=EV_ABS))
        self.assertIsNone(combination.find_analog_input_config(type_=EV_REL))
        self.assertIsNone(combination.find_analog_input_config())

    # region helper methods
    def assert_beautify_single(self, type_, code, direction, expected_beautified_name):
        """
        Assert that the beautified name of the event tuple `(type_, code,
        direction)` matches the `expected_beautified_name`.
        """
        self.assertEqual(
            InputCombination(
                InputCombination.from_tuples((type_, code, direction))
            ).beautify(),
            expected_beautified_name,
        )

    # endregion


if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-
# input-remapper - GUI for device specific keyboard mappings
# Copyright (C) 2025 sezanzeb <b8x45ygc9@mozmail.com>
#
# This file is part of input-remapper.
#
# input-remapper is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# input-remapper is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with input-remapper.  If not, see <https://www.gnu.org/licenses/>.

import asyncio
import math
import time
from functools import partial
from typing import Dict, Tuple, Optional

import evdev
from evdev.ecodes import (
    EV_REL,
    EV_ABS,
    REL_WHEEL,
    REL_HWHEEL,
    REL_WHEEL_HI_RES,
    REL_HWHEEL_HI_RES,
)

from inputremapper.configs.input_config import InputCombination, InputConfig
from inputremapper.configs.mapping import (
    Mapping,
    REL_XY_SCALING,
    WHEEL_SCALING,
    WHEEL_HI_RES_SCALING,
    DEFAULT_REL_RATE,
)
from inputremapper.injection.global_uinputs import GlobalUInputs
from inputremapper.injection.mapping_handlers.axis_transform import Transformation
from inputremapper.injection.mapping_handlers.mapping_handler import (
    MappingHandler,
    HandlerEnums,
    InputEventHandler,
)
from inputremapper.input_event import InputEvent, EventActions
from inputremapper.logging.logger import logger
from inputremapper.utils import get_evdev_constant_name


def calculate_output(value, weight, remainder):
    # self._value is between 0 and 1, scale up with weight
    scaled = value * weight + remainder
    # float_value % 1 will result in wrong calculations for negative values
    remainder = math.fmod(scaled, 1)
    return int(scaled), remainder


# TODO move into class?
async def _run_normal_output(self) -> None:
    """Start injecting events."""
    self._running = True
    self._stop = False
    remainder = 0.0
    start = time.time()

    # if the rate is configured to be slower than the default, increase the value, so
    # that the overall speed stays the same.
    rate_compensation = DEFAULT_REL_RATE / self.mapping.rel_rate
    weight = REL_XY_SCALING * rate_compensation

    while not self._stop:
        value, remainder = calculate_output(
            self._value,
            weight,
            remainder,
        )

        self._write(EV_REL, self.mapping.output_code, value)

        time_taken = time.time() - start
        sleep = max(0.0, (1 / self.mapping.rel_rate) - time_taken)
        await asyncio.sleep(sleep)
        start = time.time()

    self._running = False


# TODO move into class?
async def _run_wheel_output(self, codes: Tuple[int, int]) -> None:
    """Start injecting wheel events.

    made to inject both REL_WHEEL and REL_WHEEL_HI_RES events, because otherwise
    wheel output doesn't work for some people. See issue #354
    """
    weights = (WHEEL_SCALING, WHEEL_HI_RES_SCALING)

    self._running = True
    self._stop = False
    remainder = [0.0, 0.0]
    start = time.time()
    while not self._stop:
        for i in range(len(codes)):
            value, remainder[i] = calculate_output(
                self._value,
                weights[i],
                remainder[i],
            )

            self._write(EV_REL, codes[i], value)

        time_taken = time.time() - start
        await asyncio.sleep(max(0.0, (1 / self.mapping.rel_rate) - time_taken))
        start = time.time()

    self._running = False


class AbsToRelHandler(MappingHandler):
    """Handler which transforms an EV_ABS to EV_REL events."""

    _map_axis: InputConfig  # the InputConfig for the axis we map
    _value: float  # the current output value
    _running: bool  # if the run method is active
    _stop: bool  # if the run loop should return
    _transform: Optional[Transformation]

    def __init__(
        self,
        combination: InputCombination,
        mapping: Mapping,
        global_uinputs: GlobalUInputs,
        **_,
    ) -> None:
        super().__init__(combination, mapping, global_uinputs)

        # find the input event we are supposed to map
        assert (map_axis := combination.find_analog_input_config(type_=EV_ABS))
        self._map_axis = map_axis

        self._value = 0
        self._running = False
        self._stop = True
        self._transform = None

        # bind the correct run method
        if self.mapping.output_code in (
            REL_WHEEL,
            REL_HWHEEL,
            REL_WHEEL_HI_RES,
            REL_HWHEEL_HI_RES,
        ):
            if self.mapping.output_code in (REL_WHEEL, REL_WHEEL_HI_RES):
                codes = (REL_WHEEL, REL_WHEEL_HI_RES)
            else:
                codes = (REL_HWHEEL, REL_HWHEEL_HI_RES)

            self._run = partial(_run_wheel_output, self, codes=codes)

        else:
            self._run = partial(_run_normal_output, self)

    def __str__(self):
        name = get_evdev_constant_name(*self._map_axis.type_and_code)
        return f'AbsToRelHandler for "{name}" {self._map_axis}'

    def __repr__(self):
        return f"<{str(self)} at {hex(id(self))}>"

    @property
    def child(self):  # used for logging
        return (
            f"maps to: {self.mapping.get_output_name_constant()} "
            f"{self.mapping.get_output_type_code()} at "
            f"{self.mapping.target_uinput}"
        )

    def notify(
        self,
        event: InputEvent,
        source: evdev.InputDevice,
        suppress: bool = False,
    ) -> bool:
        if event.input_match_hash != self._map_axis.input_match_hash:
            return False

        if EventActions.recenter in event.actions:
            self._stop = True
            return True

        if not self._transform:
            absinfo = {
                entry[0]: entry[1]
                for entry in source.capabilities(absinfo=True)[EV_ABS]
            }
            self._transform = Transformation(
                max_=absinfo[event.code].max,
                min_=absinfo[event.code].min,
                deadzone=self.mapping.deadzone,
                gain=self.mapping.gain,
                expo=self.mapping.expo,
            )

        transformed = self._transform(event.value)

        self._value = transformed

        if transformed == 0:
            self._stop = True
            return True

        if not self._running:
            asyncio.ensure_future(self._run())
        return True

    def reset(self) -> None:
        self._stop = True

    def _write(self, type_, keycode, value):
        """Inject."""
        # if the mouse won't move even though correct stuff is written here,
        # the capabilities are probably wrong
        if value == 0:
            return  # rel 0 does not make sense

        try:
            self.global_uinputs.write(
                (type_, keycode, value), self.mapping.target_uinput
            )
        except OverflowError:
            # screwed up the calculation of mouse movements
            logger.error("OverflowError (%s, %s, %s)", type_, keycode, value)

    def needs_wrapping(self) -> bool:
        return len(self.input_configs) > 1

    def set_sub_handler(self, handler: InputEventHandler) -> None:
        assert False  # cannot have a sub-handler

    def wrap_with(self) -> Dict[InputCombination, HandlerEnums]:
        if self.needs_wrapping():
            return {InputCombination(self.input_configs): HandlerEnums.axisswitch}
        return {}



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# input-remapper - GUI for device specific keyboard mappings
# Copyright (C) 2025 sezanzeb <b8x45ygc9@mozmail.com>
#
# This file is part of input-remapper.
#
# input-remapper is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# input-remapper is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with input-remapper.  If not, see <https://www.gnu.org/licenses/>.

from __future__ import annotations

import asyncio

from evdev.ecodes import EV_KEY

from inputremapper.injection.macros.argument import ArgumentConfig
from inputremapper.injection.macros.macro import Macro
from inputremapper.injection.macros.task import Task


class IfSingleTask(Task):
    """If a key was pressed without combining it."""

    argument_configs = [
        ArgumentConfig(
            name="then",
            position=0,
            types=[Macro, None],
        ),
        ArgumentConfig(
            name="else",
            position=1,
            types=[Macro, None],
        ),
        ArgumentConfig(
            name="timeout",
            position=2,
            types=[int, float, None],
            default=None,
        ),
    ]

    async def run(self, callback) -> None:
        assert self.context is not None
        another_key_pressed_event = asyncio.Event()
        then = self.get_argument("then").get_value()
        else_ = self.get_argument("else").get_value()

        async def listener(event) -> None:
            if event.type != EV_KEY:
                # Ignore anything that is not a key
                return

            if event.value == 1:
                # Another key was pressed
                another_key_pressed_event.set()
                return

        self.add_event_listener(listener)

        timeout = self.get_argument("timeout").get_value()

        # Wait for anything of importance to happen, that would determine the
        # outcome of the if_single macro.
        await asyncio.wait(
            [
                asyncio.Task(another_key_pressed_event.wait()),
                asyncio.Task(self._trigger_release_event.wait()),
            ],
            timeout=timeout / 1000 if timeout else None,
            return_when=asyncio.FIRST_COMPLETED,
        )

        self.remove_event_listener(listener)

        if not self.is_holding():
            if then:
                await then.run(callback)
        else:
            # If the trigger has not been released, then `await asyncio.wait` above
            # could only have finished waiting due to a timeout, or because another
            # key was pressed.
            if else_:
                await else_.run(callback)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# input-remapper - GUI for device specific keyboard mappings
# Copyright (C) 2025 sezanzeb <b8x45ygc9@mozmail.com>
#
# This file is part of input-remapper.
#
# input-remapper is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# input-remapper is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with input-remapper.  If not, see <https://www.gnu.org/licenses/>.


import os
import tempfile
import unittest

from inputremapper.configs.paths import PathUtils
from tests.lib.test_setup import test_setup
from tests.lib.tmp import tmp


def _raise(error):
    raise error


@test_setup
class TestPaths(unittest.TestCase):
    def test_touch(self):
        with tempfile.TemporaryDirectory() as local_tmp:
            path_abcde = os.path.join(local_tmp, "a/b/c/d/e")
            PathUtils.touch(path_abcde)
            self.assertTrue(os.path.exists(path_abcde))
            self.assertTrue(os.path.isfile(path_abcde))
            self.assertRaises(
                ValueError,
                lambda: PathUtils.touch(os.path.join(local_tmp, "a/b/c/d/f/")),
            )

    def test_mkdir(self):
        with tempfile.TemporaryDirectory() as local_tmp:
            path_bcde = os.path.join(local_tmp, "b/c/d/e")
            PathUtils.mkdir(path_bcde)
            self.assertTrue(os.path.exists(path_bcde))
            self.assertTrue(os.path.isdir(path_bcde))

    def test_get_preset_path(self):
        self.assertTrue(
            PathUtils.get_preset_path().startswith(PathUtils.get_config_path())
        )
        self.assertTrue(PathUtils.get_preset_path().endswith("presets"))
        self.assertTrue(PathUtils.get_preset_path("a").endswith("presets/a"))
        self.assertTrue(
            PathUtils.get_preset_path("a", "b").endswith("presets/a/b.json")
        )

    def test_get_config_path(self):
        # might end with /beta_XXX
        self.assertTrue(
            PathUtils.get_config_path().startswith(f"{tmp}/.config/input-remapper")
        )
        self.assertTrue(PathUtils.get_config_path("a", "b").endswith("a/b"))

    def test_split_all(self):
        self.assertListEqual(PathUtils.split_all("a/b/c/d"), ["a", "b", "c", "d"])



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# input-remapper - GUI for device specific keyboard mappings
# Copyright (C) 2025 sezanzeb <b8x45ygc9@mozmail.com>
#
# This file is part of input-remapper.
#
# input-remapper is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# input-remapper is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with input-remapper.  If not, see <https://www.gnu.org/licenses/>.

"""Remove a system-wide input-remapper installation."""

import os
import sys
import shutil
import subprocess

from install.data_files import get_data_files


def uninstall() -> None:
    # remove data files
    for directory, files in get_data_files():
        for file_ in files:
            filename = os.path.basename(file_)
            path = os.path.join("/", directory, filename)

            # Removing files from system directories is risky. Low propability, very
            # high damage. To avoid accidentally removing system directories due to
            # a bug, assert that this is an input-remapper path.
            # If this happens, urgently create a new issue on github!
            assert "input" in path and "remapper" in path

            try:
                os.unlink(path)
                print("Removed", path)
            except FileNotFoundError:
                print(path, "not found")

    # language files are not in data_files
    data_path = "/usr/share/input-remapper/"
    try:
        shutil.rmtree(data_path)
        print("Removed", data_path)
    except FileNotFoundError:
        print(data_path, "not found")

    # remove pip module
    command = [
        sys.executable,
        "-m",
        "pip",
        "uninstall",
        "input-remapper",
        "--break-system-packages",
    ]
    print("Running", " ".join(command))
    # Fix the stdout ordering in github workflows
    sys.stdout.flush()
    subprocess.check_call(command)

    os.system("sudo systemctl stop input-remapper")
    os.system("sudo systemctl daemon-reload")


if __name__ == "__main__":
    uninstall()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# input-remapper - GUI for device specific keyboard mappings
# Copyright (C) 2025 sezanzeb <b8x45ygc9@mozmail.com>
#
# This file is part of input-remapper.
#
# input-remapper is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# input-remapper is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with input-remapper.  If not, see <https://www.gnu.org/licenses/>.


import os
import unittest

from inputremapper.configs.global_config import GlobalConfig
from tests.lib.test_setup import test_setup


@test_setup
class TestGlobalConfig(unittest.TestCase):
    def test_autoload(self):
        global_config = GlobalConfig()
        self.assertEqual(len(global_config.iterate_autoload_presets()), 0)
        self.assertFalse(global_config.is_autoloaded("d1", "a"))
        self.assertFalse(global_config.is_autoloaded("d2", "b"))
        self.assertEqual(global_config.get_autoload_preset("d1"), None)
        self.assertEqual(global_config.get_autoload_preset("d2"), None)

        global_config.set_autoload_preset("d1", "a")
        self.assertEqual(len(global_config.iterate_autoload_presets()), 1)
        self.assertTrue(global_config.is_autoloaded("d1", "a"))
        self.assertFalse(global_config.is_autoloaded("d2", "b"))

        global_config.set_autoload_preset("d2", "b")
        self.assertEqual(len(global_config.iterate_autoload_presets()), 2)
        self.assertTrue(global_config.is_autoloaded("d1", "a"))
        self.assertTrue(global_config.is_autoloaded("d2", "b"))
        self.assertEqual(global_config.get_autoload_preset("d1"), "a")
        self.assertEqual(global_config.get_autoload_preset("d2"), "b")

        global_config.set_autoload_preset("d2", "c")
        self.assertEqual(len(global_config.iterate_autoload_presets()), 2)
        self.assertTrue(global_config.is_autoloaded("d1", "a"))
        self.assertFalse(global_config.is_autoloaded("d2", "b"))
        self.assertTrue(global_config.is_autoloaded("d2", "c"))
        self.assertEqual(global_config._config["autoload"]["d2"], "c")
        self.assertListEqual(
            list(global_config.iterate_autoload_presets()),
            [("d1", "a"), ("d2", "c")],
        )

        global_config.set_autoload_preset("d2", None)
        self.assertTrue(global_config.is_autoloaded("d1", "a"))
        self.assertFalse(global_config.is_autoloaded("d2", "b"))
        self.assertFalse(global_config.is_autoloaded("d2", "c"))
        self.assertListEqual(
            list(global_config.iterate_autoload_presets()),
            [("d1", "a")],
        )
        self.assertEqual(global_config.get_autoload_preset("d1"), "a")

        self.assertRaises(ValueError, global_config.is_autoloaded, "d1", None)
        self.assertRaises(ValueError, global_config.is_autoloaded, None, "a")

    def test_initial(self):
        global_config = GlobalConfig()
        # when loading for the first time, create a config file with
        # the default values
        self.assertFalse(os.path.exists(global_config.path))
        global_config.load_config()
        self.assertTrue(os.path.exists(global_config.path))

        with open(global_config.path, "r") as file:
            contents = file.read()
            self.assertIn('"autoload": {}', contents)

    def test_save_load(self):
        global_config = GlobalConfig()
        self.assertEqual(len(global_config.iterate_autoload_presets()), 0)

        global_config.load_config()
        self.assertEqual(len(global_config.iterate_autoload_presets()), 0)

        global_config.set_autoload_preset("d1", "a")
        global_config.set_autoload_preset("d2", "b")

        global_config.load_config()

        self.assertListEqual(
            list(global_config.iterate_autoload_presets()),
            [("d1", "a"), ("d2", "b")],
        )


if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-
# input-remapper - GUI for device specific keyboard mappings
# Copyright (C) 2025 sezanzeb <b8x45ygc9@mozmail.com>
#
# This file is part of input-remapper.
#
# input-remapper is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# input-remapper is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with input-remapper.  If not, see <https://www.gnu.org/licenses/>.

"""Functions to assemble the mapping handler tree."""

from collections import defaultdict
from typing import Dict, List, Type, Optional, Set, Iterable, Sized, Tuple, Sequence

from evdev.ecodes import EV_KEY, EV_ABS, EV_REL

from inputremapper.configs.input_config import InputCombination, InputConfig
from inputremapper.configs.keyboard_layout import DISABLE_CODE, DISABLE_NAME
from inputremapper.configs.mapping import Mapping
from inputremapper.configs.preset import Preset
from inputremapper.exceptions import MappingParsingError
from inputremapper.injection.global_uinputs import GlobalUInputs
from inputremapper.injection.macros.parse import Parser
from inputremapper.injection.mapping_handlers.abs_to_abs_handler import AbsToAbsHandler
from inputremapper.injection.mapping_handlers.abs_to_btn_handler import AbsToBtnHandler
from inputremapper.injection.mapping_handlers.abs_to_rel_handler import AbsToRelHandler
from inputremapper.injection.mapping_handlers.axis_switch_handler import (
    AxisSwitchHandler,
)
from inputremapper.injection.mapping_handlers.combination_handler import (
    CombinationHandler,
)
from inputremapper.injection.mapping_handlers.hierarchy_handler import HierarchyHandler
from inputremapper.injection.mapping_handlers.key_handler import KeyHandler
from inputremapper.injection.mapping_handlers.macro_handler import MacroHandler
from inputremapper.injection.mapping_handlers.mapping_handler import (
    HandlerEnums,
    MappingHandler,
    ContextProtocol,
    InputEventHandler,
)
from inputremapper.injection.mapping_handlers.null_handler import NullHandler
from inputremapper.injection.mapping_handlers.rel_to_abs_handler import RelToAbsHandler
from inputremapper.injection.mapping_handlers.rel_to_btn_handler import RelToBtnHandler
from inputremapper.injection.mapping_handlers.rel_to_rel_handler import RelToRelHandler
from inputremapper.logging.logger import logger
from inputremapper.utils import get_evdev_constant_name

EventPipelines = Dict[InputConfig, Set[InputEventHandler]]

mapping_handler_classes: Dict[HandlerEnums, Optional[Type[MappingHandler]]] = {
    # all available mapping_handlers
    HandlerEnums.abs2btn: AbsToBtnHandler,
    HandlerEnums.rel2btn: RelToBtnHandler,
    HandlerEnums.macro: MacroHandler,
    HandlerEnums.key: KeyHandler,
    HandlerEnums.btn2rel: None,  # can be a macro
    HandlerEnums.rel2rel: RelToRelHandler,
    HandlerEnums.abs2rel: AbsToRelHandler,
    HandlerEnums.btn2abs: None,  # can be a macro
    HandlerEnums.rel2abs: RelToAbsHandler,
    HandlerEnums.abs2abs: AbsToAbsHandler,
    HandlerEnums.combination: CombinationHandler,
    HandlerEnums.hierarchy: HierarchyHandler,
    HandlerEnums.axisswitch: AxisSwitchHandler,
    HandlerEnums.disable: NullHandler,
}


class MappingParser:
    def __init__(
        self,
        global_uinputs: GlobalUInputs,
    ):
        self.global_uinputs = global_uinputs

    def parse_mappings(
        self,
        preset: Preset,
        context: ContextProtocol,
    ) -> EventPipelines:
        """Create a dict with a list of MappingHandler for each InputEvent."""
        handlers = []
        for mapping in preset:
            # start with the last handler in the chain, each mapping only has one output,
            # but may have multiple inputs, therefore the last handler is a good starting
            # point to assemble the pipeline
            handler_enum = self._get_output_handler(mapping)
            constructor = mapping_handler_classes[handler_enum]
            if not constructor:
                logger.warning(
                    "a mapping handler '%s' for %s is not implemented",
                    handler_enum,
                    mapping.format_name(),
                )
                continue

            output_handler = constructor(
                mapping.input_combination,
                mapping,
                context=context,
                global_uinputs=self.global_uinputs,
            )

            # layer other handlers on top until the outer handler needs ranking or can
            # directly handle a input event
            handlers.extend(self._create_event_pipeline(output_handler, context))

        # figure out which handlers need ranking and wrap them with hierarchy_handlers
        need_ranking = defaultdict(set)
        for handler in handlers.copy():
            if handler.needs_ranking():
                combination = handler.rank_by()
                if not combination:
                    raise MappingParsingError(
                        f"{type(handler).__name__} claims to need ranking but does not "
                        f"return a combination to rank by",
                        mapping_handler=handler,
                    )

                need_ranking[combination].add(handler)
                handlers.remove(handler)

        # the HierarchyHandler's might not be the starting point of the event pipeline,
        # layer other handlers on top again.
        ranked_handlers = self._create_hierarchy_handlers(need_ranking)
        for handler in ranked_handlers:
            handlers.extend(
                self._create_event_pipeline(handler, context, ignore_ranking=True)
            )

        # group all handlers by the input events they take care of. One handler might end
        # up in multiple groups if it takes care of multiple InputEvents
        event_pipelines: EventPipelines = defaultdict(set)
        for handler in handlers:
            assert handler.input_configs
            for input_config in handler.input_configs:
                logger.debug(
                    "event-pipeline with entry point: %s %s",
                    get_evdev_constant_name(*input_config.type_and_code),
                    input_config.input_match_hash,
                )
                logger.debug_mapping_handler(handler)
                event_pipelines[input_config].add(handler)

        return event_pipelines

    def _create_event_pipeline(
        self,
        handler: MappingHandler,
        context: ContextProtocol,
        ignore_ranking=False,
    ) -> List[MappingHandler]:
        """Recursively wrap a handler with other handlers until the
        outer handler needs ranking or is finished wrapping.
        """
        if not handler.needs_wrapping() or (
            handler.needs_ranking() and not ignore_ranking
        ):
            return [handler]

        handlers = []
        for combination, handler_enum in handler.wrap_with().items():
            constructor = mapping_handler_classes[handler_enum]
            if not constructor:
                raise NotImplementedError(
                    f"mapping handler {handler_enum} is not implemented"
                )

            super_handler = constructor(
                combination,
                handler.mapping,
                context=context,
                global_uinputs=self.global_uinputs,
            )
            super_handler.set_sub_handler(handler)
            for event in combination:
                # the handler now has a super_handler which takes care about the events.
                # so we need to hide them on the handler
                handler.occlude_input_event(event)

            handlers.extend(self._create_event_pipeline(super_handler, context))

        if handler.input_configs:
            # the handler was only partially wrapped,
            # we need to return it as a toplevel handler
            handlers.append(handler)

        return handlers

    def _get_output_handler(self, mapping: Mapping) -> HandlerEnums:
        """Determine the correct output handler.

        this is used as a starting point for the mapping parser
        """
        if mapping.output_code == DISABLE_CODE or mapping.output_symbol == DISABLE_NAME:
            return HandlerEnums.disable

        if mapping.output_symbol:
            if Parser.is_this_a_macro(mapping.output_symbol):
                return HandlerEnums.macro

            return HandlerEnums.key

        if mapping.output_type == EV_KEY:
            return HandlerEnums.key

        input_event = self._maps_axis(mapping.input_combination)
        if not input_event:
            raise MappingParsingError(
                f"This {mapping = } does not map to an axis, key or macro",
                mapping=Mapping,
            )

        if mapping.output_type == EV_REL:
            if input_event.type == EV_KEY:
                return HandlerEnums.btn2rel
            if input_event.type == EV_REL:
                return HandlerEnums.rel2rel
            if input_event.type == EV_ABS:
                return HandlerEnums.abs2rel

        if mapping.output_type == EV_ABS:
            if input_event.type == EV_KEY:
                return HandlerEnums.btn2abs
            if input_event.type == EV_REL:
                return HandlerEnums.rel2abs
            if input_event.type == EV_ABS:
                return HandlerEnums.abs2abs

        raise MappingParsingError(
            f"the output of {mapping = } is unknown", mapping=Mapping
        )

    def _maps_axis(self, combination: InputCombination) -> Optional[InputConfig]:
        """Whether this InputCombination contains an InputEvent that is treated as
        an axis and not a binary (key or button) event.
        """
        for event in combination:
            if event.defines_analog_input:
                return event
        return None

    def _create_hierarchy_handlers(
        self,
        handlers: Dict[InputCombination, Set[MappingHandler]],
    ) -> Set[MappingHandler]:
        """Sort handlers by input events and create Hierarchy handlers."""
        sorted_handlers = set()
        all_combinations = handlers.keys()
        events = set()

        # gather all InputEvents from all handlers
        for combination in all_combinations:
            for event in combination:
                events.add(event)

        # create a ranking for each event
        for event in events:
            # find all combinations (from handlers) which contain the event
            combinations_with_event = [
                combination for combination in all_combinations if event in combination
            ]

            if len(combinations_with_event) == 1:
                # there was only one handler containing that event return it as is
                sorted_handlers.update(handlers[combinations_with_event[0]])
                continue

            # there are multiple handler with the same event.
            # rank them and create the HierarchyHandler
            sorted_combinations = self._order_combinations(
                combinations_with_event,
                event,
            )
            sub_handlers: List[MappingHandler] = []
            for combination in sorted_combinations:
                sub_handlers.extend(handlers[combination])

            sorted_handlers.add(
                HierarchyHandler(
                    sub_handlers,
                    event,
                    self.global_uinputs,
                )
            )
            for handler in sub_handlers:
                # the handler now has a HierarchyHandler which takes care about this event.
                # so we hide need to hide it on the handler
                handler.occlude_input_event(event)

        return sorted_handlers

    def _order_combinations(
        self,
        combinations: List[InputCombination],
        common_config: InputConfig,
    ) -> List[InputCombination]:
        """Reorder the keys according to some rules.

        such that a combination a+b+c is in front of a+b which is in front of b
        for a+b+c vs. b+d+e: a+b+c would be in front of b+d+e, because the common key b
        has the higher index in the a+b+c (1), than in the b+c+d (0) list
        in this example b would be the common key
        as for combinations like a+b+c and e+d+c with the common key c: ¯\\_(ツ)_/¯

        Parameters
        ----------
        combinations
            the list which needs ordering
        common_config
            the InputConfig all InputCombination's in combinations have in common
        """
        combinations.sort(key=len)

        for start, end in self._ranges_with_constant_length(combinations.copy()):
            sub_list = combinations[start:end]
            sub_list.sort(key=lambda x: x.index(common_config))
            combinations[start:end] = sub_list

        combinations.reverse()
        return combinations

    def _ranges_with_constant_length(
        self,
        x: Sequence[Sized],
    ) -> Iterable[Tuple[int, int]]:
        """Get all ranges of x for which the elements have constant length

        Parameters
        ----------
        x: Sequence[Sized]
            l must be ordered by increasing length of elements
        """
        start_idx = 0
        last_len = 0
        for idx, y in enumerate(x):
            if len(y) > last_len and idx - start_idx > 1:
                yield start_idx, idx

            if len(y) == last_len and idx + 1 == len(x):
                yield start_idx, idx + 1

            if len(y) > last_len:
                start_idx = idx

            if len(y) < last_len:
                raise MappingParsingError(
                    "ranges_with_constant_length was called with an unordered list"
                )
            last_len = len(y)
